@@ -6,7 +6,7 @@
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
+    the Free Software Foundation; either version 3 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -59,6 +59,8 @@
 #define KEY_DOWN 2
 #define KEY_LEFT 3
 #define KEY_RIGHT 4
+#define KEY_BACKSPACE '\b'
+#define KEY_DC '\177'
 #endif
 
 #include <ctype.h>
@@ -134,13 +136,10 @@ static const char *help = N_(
 #define TAB	'\011'	/* ^I */
 #define CR	'\015'	/* ^M */
 #define ESC	'\033'	/* ^[ */
-#define DEL	'\177'
 #define CTRL_L	'\014'	/* ^L */
 #define CTRL_P	'\020'	/* ^P */
 #define CTRL_N	'\016'	/* ^N */
 #define CTRL_R	'\022'	/* ^R */
-#define DEL	'\177'
-#define BKSP 	'\b'
 
 
 #define MENUSIZE 3
@@ -176,21 +175,29 @@ struct _MenuItem {
 };
 typedef struct _MenuItem MenuItem;
 
+/* Struct for the opened device */
+struct _Context {
+	PedDevice *dev;
+	PedDisk *disk;
+	int is_devicefile;
+	char *devname;
+	char *info_size;
+	char *info_sectorsize;
+	char *info_model;
+	char *info_heads;
+	char *info_cylinders;
+	char *info_sectors;
+	char *info_cylsize;
+};
+typedef struct _Context Context;
 
-PedDevice *dev;
-PedDisk *disk;
-char *devname;
-char *info_size;
-char *info_sectorsize;
-char *info_model;
-char *info_heads;
-char *info_cylinders;
-char *info_sectors;
-char *info_cylsize;
+Context *init_disk(int new_table, char *devname, PedDevice *dev, PedDiskType *type);
 
 int new_table = 0;
 int arrow_cursor = 0;
-int is_devicefile = 0;
+
+
+
 
 
 static void
@@ -640,14 +647,15 @@ strlist_draw (const char *prompt, const StrList *strlist, int opts, const StrLis
 		y++;
 	} 
 }
-static void show_info();
-static void plist_draw (PedDisk *cdisk, PedPartition *selected, int selnum, int *start);
-static int do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartitionType havent);
+static void show_info(Context *c);
+static void plist_draw (Context *c, PedPartition *selected, int selnum, int *start);
+static int do_plist (Context *c, PedPartition **part, PedPartitionType have, PedPartitionType havent);
 
 static const StrList*
 do_strlist (const char *prompt, const StrList *strlist, int opts) {
 	const StrList *selected = strlist, *temp, *next;
-	int redraw = 1, key = 0, done = 0, start = 0, selnum = 0, i, n;
+	int redraw = 1, key = 0, done = 0, start = 0, selnum = 0, i;
+	int move_down_by;
 	const char* end_warning = opts & 1 ? 
 	                              _("No more text") : _("No more choices");
 	if (opts & 1) {
@@ -688,33 +696,35 @@ do_strlist (const char *prompt, const StrList *strlist, int opts) {
 #endif
 
 		/* We want to move 1 element when an arrow is pressed */
-		n = 1;
+		move_down_by = 1;
 		switch (key) {
 
 #if USE_KEYPAD
 			case KEY_NPAGE:
 				if (opts & 1)
-					n = LINES - SLD_HEAD - 4;
+					move_down_by = LINES - SLD_HEAD - 4;
 				else 
-					n = LINES - MENUSIZE - INFOSIZE - SLD_HEAD - 5;
+					move_down_by = LINES - MENUSIZE -
+					               INFOSIZE - SLD_HEAD - 5;
 #endif
 			case KEY_DOWN:
 				if (!selected->next) {
 					print_warning(end_warning,0);
 					break;
 				}
-				for (i = 0; i < n && selected->next; i++) {
+				for (i = 0; i < move_down_by && selected->next; i++) {
 						selected = selected->next;
 				}
-				selnum -= i;
+				selnum += i;
 				redraw = 1;
 				break;
 #if USE_KEYPAD
 			case KEY_PPAGE:
 				if (opts & 1)
-					n = LINES - SLD_HEAD - 4;
+					move_down_by = LINES - SLD_HEAD - 4;
 				else 
-					n = LINES - MENUSIZE - INFOSIZE - SLD_HEAD - 5;
+					move_down_by = LINES - MENUSIZE -
+					               INFOSIZE - SLD_HEAD - 5;
 #endif
 			case KEY_UP:
 				if (strlist == selected || 
@@ -725,13 +735,13 @@ do_strlist (const char *prompt, const StrList *strlist, int opts) {
 				
 				/* FIXME: This might be improved by editing strlist */
 				/* (next represents temp->next->next->...->next) */
-				/* We set temp and next n elements apart, and move
+				/* We set temp and next move_down_by elements apart, and move
 				   them both, until next reaches the current element,
 				   then temp points to the element n elements before it */
 				temp = strlist;
 				/* We set temp and next to n elements apart */
 				for (i = 0, next = strlist;
-				     i < n && next && next != selected;
+				     i < move_down_by && next && next != selected;
 				     i++, next = next->next);
 				/* Move them together until next is our
 				   current element, which means that temp
@@ -762,12 +772,9 @@ do_strlist (const char *prompt, const StrList *strlist, int opts) {
 					print_warning(_("Invalid key"),0);
 		}
 	}
-	if (disk) { 
-		show_info();
-		/* FIXME: This won't always print what we need, but who cares? */
-		start = 0;
-		plist_draw(disk,NULL,0,&start);
-	}
+	/* We redraw partition info */
+	show_info(NULL);
+	plist_draw(NULL, NULL, 0, 0);
 	return selected;
 }
 
@@ -808,11 +815,15 @@ read_char (const char* prompt, char **value) {
 			return 0;
 #endif
 		}
-		else if (key == BKSP || key == DEL) {
+		else if (key == KEY_BACKSPACE || 
+		         key == KEY_DC ||
+		         /* Fix for misconfigured terminals */
+		         key == '\b' ||
+		         key == '\177') {
 			if (i > 0) {
 				buf[--i] = '\0';
-				mvaddch(y,--x,' ');
-				move(y,x);
+				move(y,--x);
+				delch();
 			}
 			else
 				print_warning(_("No more characters to delete"),1);
@@ -886,11 +897,18 @@ getstring (const char* prompt, char **value, const StrList* words, const StrList
 
 static int
 getpart (const char* prompt, PedDisk **disk, PedPartition **value) {
+	/* NOTE: We should prompt the user for a disk in this case,
+	   but this will never happen in cfdisk. Also note that *disk
+	   is untouched in this function. */
 	if (!*disk) {
-		return 0; /* TODO, although it might not get used */
+		
+		return 0;
 	}
+	Context choice_context;
+	choice_context.disk = *disk;
 	menu_title(prompt);
-	return do_plist(disk,value,0,PED_PARTITION_EXTENDED | PED_PARTITION_FREESPACE);
+	return do_plist(&choice_context, value, 0, 
+	                PED_PARTITION_EXTENDED | PED_PARTITION_FREESPACE);
 }
 
 static int
@@ -1042,25 +1060,6 @@ warning_waitkey(const char *warning) {
 	notice_waitkey(warning);
 }
 
-/* Get previous non-METADATA partition. Needed for menu navigation */
-/* FIXME: Maybe it could be done faster */
-static PedPartition *
-disk_get_prev_nmd_partition(PedDisk *cdisk, PedPartition *part) {
-	PedPartition *temp_a, *temp_b;
-	/* temp_a will iterate through partitions that are not METADATA, while
-	   temp_b will iterate through all partitions. When temp_b reaches our partition,
-	   temp_a is the partition we are looking for. */
-	temp_a = NULL;
-	temp_b = temp_a;
-	do {
-		if (ped_disk_next_partition(cdisk,temp_b) == part)
-			break;
-		temp_b = ped_disk_next_partition(cdisk,temp_b);
-		if (!temp_b || !(temp_b->type & PED_PARTITION_METADATA))
-			temp_a = temp_b;
-	} while (temp_b);
-	return temp_a;
-}
 
 /* Emergency function. If some copy, move or resize fail, reread the disk from the device
    They commited it before failing, anyway. It is a bit like a hack, I might make the functions
@@ -1068,24 +1067,26 @@ disk_get_prev_nmd_partition(PedDisk *cdisk, PedPartition *part) {
  */
 /* FIXME: It would be better if we created a copy of the disk, or implemented queuing */
 static int
-_disk_reread(PedPartition **part) {
+_disk_reread(Context *c, PedPartition **part) {
 	int partnum;
 	if (*part) {
 		partnum = (*part)->num;
 	}
-	ped_disk_destroy(disk);
-	disk = ped_disk_new(dev);
+	ped_disk_destroy(c->disk);
+	c->disk = ped_disk_new(c->dev);
 	uiquery.need_commit = 0;
 	if (part) {
-		*part = ped_disk_get_partition(disk,partnum);
+		*part = ped_disk_get_partition(c->disk,partnum);
 	}
 }
 
 
 /* Here we define our menu operations */
 static int
-do_new (PedPartition **part) {
-	PedPartition *temp = disk_get_prev_nmd_partition (disk,*part);
+do_new (Context *c, PedPartition **part) {
+	char bsd_disklabel = 0x00;
+
+	PedPartition *temp = disk_get_prev_nmd_partition (c->disk,*part);
 	static MenuItem part_position[] = {
 		{ 's', N_("Begining"), N_("Create the partition at the begining of the free space") },
 		{ 'e', N_("End"), N_("Create the partition at the end of the free space") },
@@ -1096,10 +1097,23 @@ do_new (PedPartition **part) {
 	static MenuItem part_type[] = {
 		{ 'p', N_("Primary"), N_("Create primary partition") },
 		{ 'e', N_("Extended"), N_("Create extended partition") },
+		{ 'b', N_("BSD disklabel"), N_("Create a new BSD disklabel") },
 		{ ESC, N_("Cancel"), N_("Do not create a partition") },
 		{ 0, NULL, NULL }
 	};
-	const char type_keys[] = {'p', 'e', ESC, '\0'};
+	static MenuItem bsd_type[] = {
+		{ 'f', N_("FreeBSD"), N_("Create a FreeBSD partition type") },
+		{ 'o', N_("OpenBSD"), N_("Create a OpenBSD partition type") },
+		{ 'n', N_("NetBSD"), N_("Create a NetBSD partition type") },
+		{ ESC, N_("Cancel"), N_("Do not create a partition") },
+		{ 0, NULL, NULL }
+	};
+	int i = 0;
+	const char bsd_keys[] = { 'f', 'o', 'n', ESC, '\0' };
+	char type_keys[5];
+	type_keys[i++] = ESC;
+	type_keys[i++] = 'p'; /* We can always create a regular partition */
+	type_keys[i++] = 'b'; /* The BSD disklabel can be counted as a regular partition */
 	PedPartitionType type = 0;
 	PartPos pos;
 	int key,success = 1,done;
@@ -1112,42 +1126,105 @@ do_new (PedPartition **part) {
 	if((*part)->type & PED_PARTITION_LOGICAL) 
 		type = PED_PARTITION_LOGICAL;
 	else {
-		if(!can_create_primary(disk)) {
+		if(!can_create_primary(c->disk)) {
 			warning_waitkey(N_("Cannot create more primary partitions"));
 			return 0;
 		}
-		else if (can_create_extended(disk)) {
+		else if (can_create_extended(c->disk)) {
+			type_keys[i++] = 'e';
+		}
+		type_keys[i] = '\0';
 			menu_title(_("Type of partition you want to create"));
 			key = do_menu(part_type, 8, MENU_BUTTON | MENU_TITLE, type_keys, NULL);
 			if (key == ESC)	return 1;
 			else if (key == 'p') type = PED_PARTITION_NORMAL;
 			else if (key == 'e') type = PED_PARTITION_EXTENDED;
+		else if (key == 'b') {
+			type = PED_PARTITION_NORMAL;
+			key = do_menu(bsd_type, 8, MENU_BUTTON | MENU_TITLE, bsd_keys, NULL);
+			if (key == ESC) return 1;
+			else if (key == 'f') bsd_disklabel = 0xa5;
+			else if (key == 'o') bsd_disklabel = 0xa6;
+			else if (key == 'n') bsd_disklabel = 0xa9;
 		}
-		else type = PED_PARTITION_NORMAL;
+
 	}
 
 	pos.start.sector = (*part)->geom.start;
 	pos.end.sector = (*part)->geom.end;
 	if(!query_part_position(_("Where do you want to put the partition"),
 	                        part_position,&pos,pos.start.sector,
-	                        pos.end.sector,dev,NULL,&opts))
+	                        pos.end.sector,c->dev,NULL,&opts))
 		return 0;
 
-	if (!perform_mkpart(disk,&pos,type,NULL,part,opts)) {
+	PedFileSystemType *fs_type = ped_file_system_type_get("ext2");
+	if (!perform_mkpart(c->disk,&pos,type,bsd_disklabel ? fs_type : NULL,part,opts)) {
 		if (!*part) {
 			do {
-				temp = ped_disk_next_partition(disk,temp);
+				temp = ped_disk_next_partition(c->disk,temp);
 			} while (temp && temp->type & PED_PARTITION_METADATA);
 			*part = temp;
 		}
 		warning_waitkey(N_("Can't create partition."));
 		return 0;
 	}
+	if (bsd_disklabel) {
+		(*part)->fs_type = NULL;
+		set_disk_specific_system_type(*part, bsd_disklabel);
+	}
 	return 1;	
 }
 
+do_edit (Context *c, PedPartition **part) {
+	char devname[SMALLBUF];
+	PedDevice *dev;
+	if (uiquery.need_commit) {
+	        ped_exception_throw (
+				     PED_EXCEPTION_ERROR,
+				     PED_EXCEPTION_CANCEL,
+				     _("Editing a BSD label, before writing the msdos "
+			               "partition table is not supported in GNU fdisk."));
+		return 0;
+	}
+	if (!ped_device_begin_external_access(c->disk->dev))
+		return 0;
+	/* Warn if there is a filesystem on the partition */
+	if (ped_file_system_probe(&((*part)->geom))) {
+		if (ped_exception_throw(PED_EXCEPTION_WARNING,
+					PED_EXCEPTION_YES_NO,
+		     _("This partition seems to contain a filesystem. "
+		       "This is going to destroy it. "
+		       "Are you sure you want to continue?")) 
+			== PED_EXCEPTION_NO) {
+			ped_device_end_external_access(c->disk->dev);
+			return 0;
+		}
+	}
+	/* If there is no already created BSD label created,
+	   we warn the user about it */
+	else if (!is_bsd_partition(c->disk->dev->path, 
+		(*part)->geom.start * c->disk->dev->sector_size, 
+		c->disk->dev->sector_size)) {
+		if (ped_exception_throw(PED_EXCEPTION_WARNING,
+					PED_EXCEPTION_YES_NO,
+		     _("There is no BSD label on this partition. "
+		       "Do you want to create one?")) 
+			== PED_EXCEPTION_NO) {
+			ped_device_end_external_access(c->disk->dev);
+			return 0;
+		}
+	}
+	get_partition_device(devname, SMALLBUF, *part, 0);
+	Context * bsd_context = init_disk(0, devname, NULL, ped_disk_type_get ("bsd"));
+	show_info(bsd_context);
+	do_plist(bsd_context,NULL,0,0);
+
+	/* We reset this back to its default state */
+	uiquery.need_commit = 0;
+	ped_device_end_external_access(c->disk->dev);
+}
 static int
-do_resize (PedPartition **part) {
+do_resize (Context *c, PedPartition **part) {
 	static MenuItem part_position[] = {
 		{ 's', N_("Fixed start"), N_("Don't move the beginning of the partition") },
 		/* FIXME: Wording. */
@@ -1225,7 +1302,7 @@ do_resize (PedPartition **part) {
 		last = (*part)->geom.end;
 
 	if(!query_part_position(_("Where to place the resized partition"),
-	                        part_position,&pos,first,last,dev,
+	                        part_position,&pos,first,last,c->dev,
 	                        constraint,&opts)) {
 		if (constraint) ped_constraint_destroy (constraint);
 		return 0;
@@ -1235,18 +1312,18 @@ do_resize (PedPartition **part) {
 
 	if (opts & UI_CUSTOM_VALUES || pos.start.sector != (*part)->geom.start 
 			|| pos.end.sector != (*part)->geom.end)
-		if (!perform_resize(disk,*part,&pos,opts)) {
+		if (!perform_resize(c->disk,*part,&pos,opts)) {
 			warning_waitkey(N_("Couldn't resize partition."));
 			/* Ooopsy */
 			if (!(opts & UI_NO_FS_RESIZE) && uiquery.need_commit)
-				_disk_reread(part);
+				_disk_reread(c, part);
 			return 0;
 		}
 	return 1;
 }
 
 static int
-do_move (PedPartition **part) {
+do_move (Context *c, PedPartition **part) {
 	/* Combine this with new partition? */
 	static MenuItem part_position[] = {
 		{ 's', N_("Begining"), N_("Move the partition to the begining of the free space") },
@@ -1266,14 +1343,14 @@ do_move (PedPartition **part) {
 		warning_waitkey(N_("Can't open the filesystem"));
 		return 0;
 	}
-	constraint = ped_file_system_get_copy_constraint(fs,dev);
+	constraint = ped_file_system_get_copy_constraint(fs,c->dev);
 	if (!constraint) {
 		warning_waitkey(N_("We can't move this partition"));
 		return 0;
 	}
 	ped_file_system_close (fs);
 	menu_title(_("Select free space to move the partition to"));
-	if (!do_plist(&disk, &dst, PED_PARTITION_FREESPACE, 0)) {
+	if (!do_plist(c, &dst, PED_PARTITION_FREESPACE, 0)) {
 		ped_constraint_destroy (constraint);
 		return 0;
 	}
@@ -1297,22 +1374,22 @@ do_move (PedPartition **part) {
 	pos.end.sector = end;
 	if(!query_part_position(_("Where do you want to move the partition"),
 	                        part_position,&pos,start,dst->geom.end,
-	                        dev,constraint,&opts)) {
+	                        c->dev,constraint,&opts)) {
 		ped_constraint_destroy (constraint);
 		return 0;
 	}
 	ped_constraint_destroy (constraint);
-	if (!perform_move(disk,*part,&pos,opts)) {
+	if (!perform_move(c->disk,*part,&pos,opts)) {
 		warning_waitkey(N_("Partition move failed"));
 		/* Oooops */
-		if (uiquery.need_commit) _disk_reread(part);
+		if (uiquery.need_commit) _disk_reread(c, part);
 		return 0;
 	}
 	return 1;
 }
 
 static int
-do_mkfs (PedPartition **part) {
+do_mkfs (Context *c, PedPartition **part) {
 	char buf[SMALLBUF];
 	const PedFileSystemType *type = NULL;
 	int go;
@@ -1330,7 +1407,7 @@ do_mkfs (PedPartition **part) {
 			warning_waitkey(buf);
 		}
 	}
-	if (!perform_mkfs (disk, *part, type, UI_WARN_COMMIT)) {
+	if (!perform_mkfs (c->disk, *part, type, UI_WARN_COMMIT)) {
 		warning_waitkey(N_("The filesystem was not created successfully"));
 		return 0;
 	}
@@ -1339,8 +1416,8 @@ do_mkfs (PedPartition **part) {
 }
 
 static int
-do_check (PedPartition **part) {
-	if (!perform_check(disk,*part)) {
+do_check (Context *c, PedPartition **part) {
+	if (!perform_check(c->disk,*part)) {
 		warning_waitkey(N_("The consistency check failed."));
 		return 0;
 	}
@@ -1349,7 +1426,7 @@ do_check (PedPartition **part) {
 }
 
 static int
-do_rescue(PedPartition **part) {
+do_rescue(Context *c, PedPartition **part) {
 	static MenuItem rescue_menu[] = {
 		{ 'g', N_("Continue"), N_("Proceed with the rescue") },
 		{ 'c', N_("Custom"), N_("Select custom area to look for partitions (for experts only)") },
@@ -1370,7 +1447,7 @@ do_rescue(PedPartition **part) {
 	if (temp && temp->type & PED_PARTITION_FREESPACE)
 		temp = disk_get_prev_nmd_partition(disk, temp); 
 #endif
-	if(!perform_rescue (disk, (*part)->geom.start, (*part)->geom.end, 
+	if(!perform_rescue (c->disk, (*part)->geom.start, (*part)->geom.end, 
 	           UI_WARN_COMMIT | (key == 'c' ? UI_CUSTOM_VALUES : 0))) {
 		warning_waitkey(N_("There was an error during rescue"));
 		return 0;
@@ -1383,16 +1460,16 @@ do_rescue(PedPartition **part) {
 #endif
 	*part = NULL;
 	do {
-		*part = ped_disk_next_partition(disk,*part);
+		*part = ped_disk_next_partition(c->disk,*part);
 	} while (*part && (*part)->type & PED_PARTITION_METADATA);
 	notice_waitkey(N_("Finished looking for partitions"));
 	return 1;
 }
 
 static int
-do_copy(PedPartition **part) {
-	if (!perform_cp(disk,*part, UI_WARN_COMMIT)) {
-		if (uiquery.need_commit) _disk_reread(part);
+do_copy(Context *c, PedPartition **part) {
+	if (!perform_cp(c->disk,*part, UI_WARN_COMMIT)) {
+		if (uiquery.need_commit) _disk_reread(c, part);
 		warning_waitkey(N_("Partition not copied successfully"));
 		return 0;
 	}
@@ -1401,35 +1478,35 @@ do_copy(PedPartition **part) {
 }
 
 static int
-do_delete(PedPartition **part) {
+do_delete(Context *c, PedPartition **part) {
 	int go = 1;
 	PedPartition *temp;
 	/* Don't delete the first partition on a mac partition table */
-	if (!strcmp(disk->type->name,"mac") && (*part)->num == 1) {
+	if (!strcmp(c->disk->type->name,"mac") && (*part)->num == 1) {
 		warning_waitkey(N_("You should not delete this partition."));
 		return 0;
 	}
 	/* Remember either the previous partition so we can select the free space */
-	temp = disk_get_prev_nmd_partition(disk, *part);
+	temp = disk_get_prev_nmd_partition(c->disk, *part);
 	if (temp && temp->type & PED_PARTITION_FREESPACE)
-		temp = disk_get_prev_nmd_partition(disk, temp); 
+		temp = disk_get_prev_nmd_partition(c->disk, temp); 
 
 	getbool(_("Do you want to delete this partition?"),&go);
 	if (!go) return 0;
-	if (!perform_rm(disk,*part)) {
+	if (!perform_rm(c->disk,*part)) {
 		warning_waitkey(N_("Can't delete partition"));
 		return 0;
 	}
 	do {
-		temp = ped_disk_next_partition(disk,temp);
+		temp = ped_disk_next_partition(c->disk,temp);
 	} while (temp && temp->type & PED_PARTITION_METADATA); 
 	*part = temp;
 	return 1;
 }
 
 static int
-do_maximize(PedPartition **part) {
-	if (!perform_maximize(disk,*part)) {
+do_maximize(Context *c, PedPartition **part) {
+	if (!perform_maximize(c->disk,*part)) {
 		warning_waitkey(N_("Couldn't maximize this partition"));
 		return 0;
 	}
@@ -1437,22 +1514,22 @@ do_maximize(PedPartition **part) {
 }
 
 static int
-do_minimize(PedPartition **part) {
+do_minimize(Context *c, PedPartition **part) {
 	PedPartition *temp;
-	temp = disk_get_prev_nmd_partition(disk, *part);
+	temp = disk_get_prev_nmd_partition(c->disk, *part);
 	if (temp && temp->type & PED_PARTITION_FREESPACE)
-		temp = disk_get_prev_nmd_partition(disk, temp); 
+		temp = disk_get_prev_nmd_partition(c->disk, temp); 
 
 	if (!(*part)->type & PED_PARTITION_EXTENDED)
 		return 0;
-	if (!ped_disk_minimize_extended_partition(disk)) {
+	if (!ped_disk_minimize_extended_partition(c->disk)) {
 		warning_waitkey(N_("Couldn't minimize the extended partition"));
 		return 0;
 	}
-	if (!ped_disk_extended_partition(disk)) {
+	if (!ped_disk_extended_partition(c->disk)) {
 		*part = temp;
 		do {
-			temp = ped_disk_next_partition(disk,temp);
+			temp = ped_disk_next_partition(c->disk,temp);
 		} while (temp && temp->type & PED_PARTITION_METADATA);
 		if (temp) *part = temp;
 	}
@@ -1460,8 +1537,8 @@ do_minimize(PedPartition **part) {
 }
 
 static int
-do_commit () {
-	if (!perform_commit(disk,UI_WARN_COMMIT)) {
+do_commit (Context *c) {
+	if (!perform_commit(c->disk,UI_WARN_COMMIT)) {
 		warning_waitkey(N_("Commit failed."));
 		return 0;
 	} 
@@ -1515,7 +1592,7 @@ do_units () {
 /* I'm not quite sure we should let the user do this... */
 /* Actually, I'm still not quite sure if it does what it says to do :) */
 static int
-do_type (PedPartition **part) {
+do_type (Context *c, PedPartition **part) {
 	static MenuItem type_menu[] = {
 		{ 'a', N_("Auto"), N_("Try to automatically set the correct type") },
 		{ 'c', N_("Custom"), N_("Select a custom filesystem type (for experts only)") },
@@ -1532,7 +1609,7 @@ do_type (PedPartition **part) {
 		return 0;
 	else if (key == 'a')
 		type = ped_file_system_probe(&(*part)->geom);
-	else if (!perform_set_system(disk,*part,type)) {
+	else if (!perform_set_system(c->disk,*part,type)) {
 		warning_waitkey(N_("Couldn't change the filesystem type"));
 		return 0;
 	}
@@ -1540,12 +1617,12 @@ do_type (PedPartition **part) {
 }
 
 static int
-do_name (PedPartition **part) {
-	if (!ped_disk_type_check_feature(disk->type,PED_DISK_TYPE_PARTITION_NAME)) {
+do_name (Context *c, PedPartition **part) {
+	if (!ped_disk_type_check_feature(c->disk->type,PED_DISK_TYPE_PARTITION_NAME)) {
 		warning_waitkey(N_("The partition label doesn't support partition names"));
 		return 0;
 	}
-	if (!perform_name(disk,*part, NULL)) {
+	if (!perform_name(c->disk,*part, NULL)) {
 		warning_waitkey(N_("Name wasn't changed successfully."));
 		return 0;
 	} 
@@ -1643,7 +1720,7 @@ flag_draw (const PedPartition *part, PedPartitionFlag selected) {
 }
 
 static int
-do_flag (PedPartition **part) {
+do_flag (Context *c, PedPartition **part) {
 	PedPartitionFlag walk;
 	/* We will put the flags in this array */
 	PedPartitionFlag flags[MAXFLAGS+1];
@@ -1720,7 +1797,7 @@ do_flag (PedPartition **part) {
 			}
 		}
 		else if (key == CR || key == ' ') {
-			perform_set(disk,*part,flags[i],UI_FLAG_TOGGLE);
+			perform_set(c->disk,*part,flags[i],UI_FLAG_TOGGLE);
 			redraw = 1;
 		}
 		else 
@@ -1738,9 +1815,9 @@ do_help () {
 }
 
 static int
-do_pinfo (PedPartition **part) {
+do_pinfo (Context *c, PedPartition **part) {
 	StrList *info = NULL;
-	char buf[SMALLBUF];
+	char buf[SMALLBUF], buf2[SMALLBUF];
 	const char *temp;
 	int n;
 
@@ -1750,11 +1827,12 @@ do_pinfo (PedPartition **part) {
 
 	/* TODO: I did like it is done in Linux cfdisk. 
 	         This won't be always correct */
-	n = strlen(dev->path);
-	if (!is_devicefile && !((*part)->type & PED_PARTITION_FREESPACE)) {
-		snprintf(buf,SMALLBUF,
-			isdigit(dev->path[n-1]) ? "%30s: %sp%d" : "%30s: %s%d",
-			_("Possible partition device"), dev->path, (*part)->num);
+	n = strlen(c->dev->path);
+	if (!(0 && c->is_devicefile) && !((*part)->type & PED_PARTITION_FREESPACE)) {
+		
+		get_partition_device(buf2, SMALLBUF, *part, 0);
+		snprintf(buf,SMALLBUF, "%30s: %s",
+			_("Possible partition device"), buf2);
 		info = str_list_append(info,buf);
 	}
 	if ((*part)->type & PED_PARTITION_FREESPACE) {
@@ -1771,7 +1849,7 @@ do_pinfo (PedPartition **part) {
 		temp = _("Primary");
 	snprintf(buf,SMALLBUF,"%30s: %s", _("Partition type"), temp);
 	info = str_list_append(info,buf);
-	if (ped_disk_type_check_feature(disk->type,PED_DISK_TYPE_PARTITION_NAME)) {
+	if (ped_disk_type_check_feature(c->disk->type,PED_DISK_TYPE_PARTITION_NAME)) {
 		snprintf(buf,SMALLBUF,"%30s: %s", _("Partition name"), ped_partition_get_name(*part));
 		info = str_list_append(info,buf);
 	}
@@ -1835,6 +1913,14 @@ do_pinfo (PedPartition **part) {
 	return 1;
 }
 
+/* We remember last disk listed here */
+static struct {
+	Context *c;
+	PedPartition *selected;
+	int selnum;
+	int start;
+} last_plist = { NULL, NULL, 0, 0 };
+
 /* Define column positions. The first three will be fixed. */
 #define col_name 3
 #define col_flags 12
@@ -1845,7 +1931,22 @@ do_pinfo (PedPartition **part) {
 
 /* Partition list drawing function */
 static void
-plist_draw (PedDisk *cdisk, PedPartition *selected, int selnum, int *start) {
+plist_draw (Context *c, PedPartition *selected, int selnum, int *start) {
+	if (!c) {
+		c = last_plist.c;
+		selected = last_plist.selected;
+		selnum = last_plist.selnum;
+		start = &last_plist.start;
+	}
+	else {
+		last_plist.c = c;
+		last_plist.selected = selected;
+		last_plist.selnum = selnum;
+		last_plist.start = *start;
+	}
+	if (!c || !c->disk) {
+		return;
+	}
 	PedPartition *part;
 	const char* temp;
 	char buf[SMALLBUF],fsbuf[FSBUF];
@@ -1854,10 +1955,10 @@ plist_draw (PedDisk *cdisk, PedPartition *selected, int selnum, int *start) {
 	int col_size = ((double)(66-col_fs)/(80-col_fs))*(COLS-col_fs)+col_fs;
 	int n,i,x,y = INFOSIZE;
 
-	int can_primary = can_create_primary(disk);
-	int can_extended = can_create_extended(disk);
+	int can_primary = can_create_primary(c->disk);
+	int can_extended = can_create_extended(c->disk);
 	/* TODO: Should we make can_name global? */
-	int can_name = ped_disk_type_check_feature(cdisk->type,PED_DISK_TYPE_PARTITION_NAME);
+	int can_name = ped_disk_type_check_feature(c->disk->type,PED_DISK_TYPE_PARTITION_NAME);
 
 	n = LINES-3-MENUSIZE;
 
@@ -1890,8 +1991,8 @@ plist_draw (PedDisk *cdisk, PedPartition *selected, int selnum, int *start) {
 	mvaddstr(y++,0,buf);
 
 	n = LINES-3-MENUSIZE;
-	for (part = ped_disk_next_partition (cdisk, NULL), i=0; part && y < n;
-	     part = ped_disk_next_partition (cdisk, part)) {
+	for (part = ped_disk_next_partition (c->disk, NULL), i=0; part && y < n;
+	     part = ped_disk_next_partition (c->disk, part)) {
 		if (part->type & PED_PARTITION_METADATA) continue; /* We skip METADATA partitions */
 
 		i++;
@@ -1983,6 +2084,7 @@ plist_draw (PedDisk *cdisk, PedPartition *selected, int selnum, int *start) {
 static MenuItem main_menu[] = {
 	{ 'b', N_("Flags"), N_("Change the flags of the current partition") },
 	{ 'n', N_("New"), N_("Create new partition from free space") },
+	{ 'e', N_("Edit"), N_("Edit this BSD disklabel") },
 	{ 's', N_("Rescue"), N_("Look for deleted and corrupted partitions in the free space") },
 	{ 'f', N_("Make FS"), N_("Creates a filesystem on the partition") },
 	{ 'c', N_("Check"), N_("Check partition for consistency") },
@@ -1996,7 +2098,7 @@ static MenuItem main_menu[] = {
 	{ 't', N_("Type"), N_("Set the filesystem type (doesn't convert the filesystem)") },
 	{ 'u', N_("Units"), N_("Change units of the partition size display") },
 	{ 'w', N_("Commit"), N_("Write the changes to the disk") },
-	{ 'q', N_("Quit"), N_("Quit program without writing partition table") },
+	{ 'q', N_("Quit"), N_("End editing this partition table") },
 	{ 'i', N_("Info"), N_("Display additional partition information") },
 	{ 'h', N_("Help"), N_("Display help") },
         { 0, NULL, NULL }
@@ -2006,6 +2108,7 @@ const char keys_ext[] = { 'm','d','u','w','q','i','h','r','x','z', ESC, '\0' };
 const char keys_free[] = { 'n','s','u','w','q','h','i', ESC, '\0' };
 const char keys_part[] = { 'b','f','c','m','y','r','o','d','t','u','w','q','i','h', ESC, '\0' };
 /*bfcayrodtuwqph*/
+const char keys_bsd[] = { 'b', 'e', 'd', 't', 'u', 'w', 'q', 'i', 'h', ESC, '\0' };
 
 static MenuItem partselect_menu[] = {
 	{ 's', N_("Select"), N_("Select this as the source partition") },
@@ -2019,7 +2122,7 @@ const char keys_cantselect[] = { ESC, '\0' };
 /* We need to separate this to use the partition list for other purposes like partition copy */
 /* FIXME Make use of redraw or get rid of it completely */
 static int
-main_plist(PedPartition **part, int key) {
+main_plist(Context *c, PedPartition **part, int key) {
 	switch (key) {
 		case ESC:
 		case 'q':
@@ -2034,22 +2137,27 @@ main_plist(PedPartition **part, int key) {
 			}
 			break;
 		case 'n':
-			do_new(part);
+			do_new(c, part);
+			//redraw = 1;
+			key = 0;
+			break;
+		case 'e':
+			do_edit(c, part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'w':
-			do_commit();
+			do_commit(c);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'f':
-			do_mkfs(part);
+			do_mkfs(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'c':
-			do_check(part);
+			do_check(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
@@ -2059,47 +2167,47 @@ main_plist(PedPartition **part, int key) {
 			key = 0;
 			break;
 		case 's':
-			do_rescue(part);
+			do_rescue(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'y':
-			do_copy(part);
+			do_copy(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'r':
-			do_resize(part);
+			do_resize(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'd':
-			do_delete(part);
+			do_delete(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'o':
-			do_move(part);
+			do_move(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'x':
-			do_maximize(part);
+			do_maximize(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'z':
-			do_minimize(part);
+			do_minimize(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 'b':
-			do_flag(part);
+			do_flag(c,part);
 			//redraw = 1;
 			key = 0;
 			break;
 		case 't':
-			do_type(part);
+			do_type(c,part);
 			//redraw = 0; //??
 			key = 0;
 			break;
@@ -2109,12 +2217,12 @@ main_plist(PedPartition **part, int key) {
 			key = 0;
 			break;
 		case 'm':
-			do_name(part);
+			do_name(c,part);
 			//redraw = 0;
 			key = 0;
 			break;
 		case 'i':
-			do_pinfo(part);
+			do_pinfo(c,part);
 			//redraw = 0;
 			key = 0;
 			break;
@@ -2130,7 +2238,7 @@ main_plist(PedPartition **part, int key) {
 /* Okay, this is our partition list. If there is no pointer to partition given, ASSume main menu
    have and havent aren't perfect, but work for our purposes */
 static int
-do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartitionType havent) {
+do_plist (Context *c, PedPartition **part, PedPartitionType have, PedPartitionType havent) {
 	int key = 0, i, redraw = 1, selnum = 0, start = 0;
 	const char* keys;
 	PedPartition *selected = NULL;
@@ -2145,12 +2253,12 @@ do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartit
 		menu_opts |= MENU_TITLE;
 	}
 	do {
-		selected = ped_disk_next_partition(*cdisk,selected);
+		selected = ped_disk_next_partition(c->disk,selected);
 	} while (selected && selected->type & PED_PARTITION_METADATA);
 	while (!key) {
 		if(redraw) {
-			show_info();
-			plist_draw(*cdisk,selected, selnum, &start);
+			show_info(c);
+			plist_draw(c,selected, selnum, &start);
 			redraw = 0;
 		}
 		if (part) {
@@ -2163,13 +2271,15 @@ do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartit
 			keys = keys_ext;
 		else if (selected->type & PED_PARTITION_FREESPACE)
 			keys = keys_free;
+		else if (is_part_type_bsd(selected))
+			keys = keys_bsd;
 		else keys = keys_part;
 		
 		key = do_menu(part_menu, 8, menu_opts, keys, &menupos);
 		menu_opts &= ~MENU_LEAVE_WARNING;
 		if (key == KEY_UP) {
 
-			temp = disk_get_prev_nmd_partition(*cdisk,selected);
+			temp = disk_get_prev_nmd_partition(c->disk,selected);
 			if (temp) {
 				selected = temp;
 				selnum--;
@@ -2183,9 +2293,9 @@ do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartit
 			key = 0;
 		}
 		else if (key == KEY_DOWN) {
-			temp = ped_disk_next_partition(*cdisk,selected);
+			temp = ped_disk_next_partition(c->disk,selected);
 			while (temp && temp->type & PED_PARTITION_METADATA)
-				temp = ped_disk_next_partition(*cdisk,temp);
+				temp = ped_disk_next_partition(c->disk,temp);
 			if (temp) { 
 				selnum++;
 				selected = temp;
@@ -2208,13 +2318,13 @@ do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartit
 				}
 				key = 0;
 			} else {
-				key = main_plist(&selected, key);
+				key = main_plist(c, &selected, key);
 				/* Count which partition is now selected */
 				PedPartition *temp;
 				selnum = 0;
-				for(temp = ped_disk_next_partition(*cdisk,NULL);
+				for(temp = ped_disk_next_partition(c->disk,NULL);
 				    temp && temp != selected;
-				    temp = ped_disk_next_partition(*cdisk,temp))
+				    temp = ped_disk_next_partition(c->disk,temp))
 					selnum++;
 				redraw = 1;
 			}
@@ -2230,7 +2340,13 @@ do_plist (PedDisk **cdisk, PedPartition **part, PedPartitionType have, PedPartit
 
 /* This prints the information at the top */
 static void
-show_info() {
+show_info(Context *c) {
+	if (!c) {
+		c = last_plist.c;
+		if (!c) {
+			return;
+		}
+	}
 	int y;
 	char buf[SMALLBUF];
 	for (y = 0; y < INFOSIZE+1; y++) {
@@ -2242,21 +2358,21 @@ show_info() {
 	mvaddstr(y++,get_center(buf),buf);
 	y++;
 	
-	if (info_model) 
-		mvaddstr(y++,get_center(info_model),info_model);
+	if (c->info_model) 
+		mvaddstr(y++,get_center(c->info_model),c->info_model);
 	
-	if (info_size) {
+	if (c->info_size) {
 		snprintf(buf,SMALLBUF,_("Disk: %s   Disk type: %s    Size: %s"), 
-		         devname, disk->type->name, info_size);
+		         c->devname, c->disk->type->name, c->info_size);
 	}
 	else {
-		snprintf(buf,SMALLBUF,_("Disk: %s   Disk type: %s"), devname, disk->type->name);
+		snprintf(buf,SMALLBUF,_("Disk: %s   Disk type: %s"), c->devname, c->disk->type->name);
 	}
 	mvaddstr(y++,get_center(buf),buf);
 	
-	if (info_heads && info_sectors && info_cylinders) {
+	if (c->info_heads && c->info_sectors && c->info_cylinders) {
 		snprintf(buf,SMALLBUF, _("Heads: %s   Sectors per track: %s   Cylinders: %s"),
-				info_heads, info_sectors, info_cylinders);
+				c->info_heads, c->info_sectors, c->info_cylinders);
 		mvaddstr(y++,get_center(buf),buf);
 	}
 	
@@ -2266,10 +2382,21 @@ show_info() {
 
 
 
-static void
-init_disk() {
+Context *
+init_disk(int new_table, char *devname, PedDevice *dev, PedDiskType *type) {
 	char buf[SMALLBUF];
 	int n;
+	PedDisk *disk = NULL;
+	int is_devicefile = 0;
+	/*
+	char *info_size;
+	char *info_sectorsize;
+	char *info_model;
+	char *info_heads;
+	char *info_cylinders;
+	char *info_sectors;
+	char *info_cylsize;
+	*/
 	static MenuItem custom_part[] = {
 		{ 'c', N_("Choose"), N_("Prompt for device address") },
 		{ 'q', N_("Quit"), N_("Quit program") },
@@ -2283,7 +2410,13 @@ init_disk() {
 	/* If we can't find the device automatically. FIXME: This is not tested. */
 	if (!dev) {
 		menu_title(_("Cannot find a device automatically"));
-		if ('c' == do_menu(custom_part, 8, MENU_BUTTON | MENU_TITLE, "cq", 0)) {
+		if (devname) {
+			dev = ped_device_get(devname);
+			if (!dev || !ped_device_open(dev)) {
+				do_quit(1,_("Invalid device"));
+			}
+		}
+		else if ('c' == do_menu(custom_part, 8, MENU_BUTTON | MENU_TITLE, "cq", 0)) {
 			read_char(_("Enter path to the device"),&devname);
 			if (!devname) {
 				do_quit(0,NULL);
@@ -2304,8 +2437,8 @@ init_disk() {
 			ped_exception_catch();
 			ped_exception_leave_all();
 			menu_title(_("Cannot find a partition table on the disk"));
-			if ('c' == do_menu(custom_label, 8, MENU_BUTTON | MENU_TITLE, "cq", 0)) {
-				if (!perform_mklabel (dev, &disk, NULL))
+			if (type || 'c' == do_menu(custom_label, 8, MENU_BUTTON | MENU_TITLE, "cq", 0)) {
+				if (!perform_mklabel (dev, &disk, type))
 					do_quit(1,_("Creation of partition table failed"));
 			}
 			else do_quit (0,NULL);
@@ -2316,53 +2449,68 @@ init_disk() {
 		if (!perform_mklabel (dev, &disk, NULL))
 			do_quit(1,_("Creation of partition table failed"));
 	}
+
 	//ped_unit_set_default(PED_UNIT_MEGABYTE);
+
+	Context *context = malloc(sizeof(Context));
+
+	context->dev = dev;
+	context->disk = disk;
+	context->is_devicefile = is_devicefile;
+	context->devname = devname;
+
+	
 	/* Initialize the info variables */
 	snprintf(buf,SMALLBUF,_("%s, %s"),
 		ped_unit_format_custom_byte (dev, dev->length * dev->sector_size-1, PED_UNIT_BYTE),
 		ped_unit_format_custom_byte (dev, dev->length * dev->sector_size-1, PED_UNIT_MEGABYTE));
 	n = strlen(buf);
-	info_size = calloc(n,sizeof(char));
-	strncpy(info_size,buf,n);
+	context->info_size = calloc(n,sizeof(char));
+	strncpy(context->info_size,buf,n);
 
-	/* FIXME: Isn't changing the pointer enough, why we cpy? */
-	if (is_devicefile) 
-		info_model = _("Device file");
+	if (is_devicefile) {
+		char * model_file = _("Device file");
+		context->info_model = calloc(n,sizeof(char));
+		strncpy(context->info_model,model_file,n);
+	}
 	else {
 		n = strlen(dev->model);
-		info_model = calloc(n,sizeof(char));
-		strncpy(info_model,dev->model,n);
+		context->info_model = calloc(n,sizeof(char));
+		strncpy(context->info_model,dev->model,n);
 	}
 	
 	snprintf(buf,SMALLBUF,"%lldB/%lldB\n",dev->sector_size, dev->phys_sector_size);
 	n = strnlen(buf);
-	info_sectorsize = calloc(n,sizeof(char));
-	strncpy(info_sectorsize,buf,n);
+	context->info_sectorsize = calloc(n,sizeof(char));
+	strncpy(context->info_sectorsize,buf,n);
 
 	PedCHSGeometry* chs = &(dev->bios_geom);
-        info_cylsize = ped_unit_format_custom (dev, chs->heads * chs->sectors, PED_UNIT_KILOBYTE);
+        context->info_cylsize = ped_unit_format_custom (dev, 
+			chs->heads * chs->sectors, PED_UNIT_KILOBYTE);
 
 	snprintf(buf, SMALLBUF, "%d", chs->cylinders);
 	n = strlen(buf);
-	info_cylinders = calloc(n,sizeof(char));
-	strncpy(info_cylinders,buf,n);
+	context->info_cylinders = calloc(n,sizeof(char));
+	strncpy(context->info_cylinders,buf,n);
 
 	snprintf(buf, SMALLBUF, "%d", chs->heads);
 	n = strlen(buf);
-	info_heads = calloc(n,sizeof(char));
-	strncpy(info_heads,buf,n);
+	context->info_heads = calloc(n,sizeof(char));
+	strncpy(context->info_heads,buf,n);
 	
 	snprintf(buf, SMALLBUF, "%d",  chs->sectors);
 	n = strlen(buf);
-	info_sectors = calloc(n,sizeof(char));
-	strncpy(info_sectors,buf,n);
+	context->info_sectors = calloc(n,sizeof(char));
+	strncpy(context->info_sectors,buf,n);
+	
+	return context;
 
 }
 
 
 
 static void
-do_ui (void) {
+do_ui (char *devname, PedDevice *dev) {
 	initscr();
 	/* TODO: Maybe we should enable colouring? */
 	//start_color();
@@ -2374,10 +2522,10 @@ do_ui (void) {
 	nonl();
 
 	init_calls();
-	init_disk();
 
-	show_info();
-	do_plist(&disk,NULL,0,0);
+	Context * global_context = init_disk(new_table, devname, dev, NULL);
+	show_info(global_context);
+	do_plist(global_context,NULL,0,0);
 	endwin();
 }
 
@@ -2409,6 +2557,8 @@ print_version() {
 int
 main (int argc, char **argv) {
 	PedUnit unit = -1;
+	char *devname;
+	PedDevice *dev;
 #ifdef HAVE_GETOPT_H
 	static struct option long_options[] = {
 		{ "new-table", no_argument, NULL, 'z' },
@@ -2470,5 +2620,5 @@ main (int argc, char **argv) {
 		printf(_("Invalid device\n"));
 		exit(1);
 	}
-	do_ui();
+	do_ui(devname, dev);
 }
